/* File: gosiv.h

* Copyright (c) 2003 James A. Strother
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

#ifndef _GOSIV_H
#define _GOSIV_H

#include <math.h>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
using namespace std;

#include "gosiv_types.h"
#include "gosiv_path.h"

class gosiv
{
  enum
    {
      GOSIV_BAD,
      GOSIV_PS,
      GOSIV_SVG
    };

  enum
    {
      GOSIV_MOVETO,
      GOSIV_LINETO,
      GOSIV_CURVETO,
      GOSIV_CLOSE
  };

  struct gosiv_ps
  {
    string    fn;   /* filename    */
    ofstream  fs;   /* file stream */
    ostream*  os;   /* out stream  */
  };

  struct gosiv_svg
  {
    string    fn;   /* filename    */
    ofstream  fs;   /* file stream */
    ostream*  os;   /* out stream  */
  };
  
 public:
  gosiv();
  ~gosiv();

  void openPS(const gosiv_canvas&,const string&);
  void openSVG(const gosiv_canvas&,const string&);

  bool bad();
  void close();

  void drawBack(const gosiv_back&);
  void drawArrow(const gosiv_arrow&);
  void drawLegend(const gosiv_legend&);

 private:
  int   type;
  void* target;

  gosiv_canvas canvas;

  void drawBackgroundSolid(const gosiv_back&);
  void drawArrowTailLine(const gosiv_arrow&);
  void drawArrowTailBlock(const gosiv_arrow&);
  void drawArrowHeadOutline(const gosiv_arrow&);
  void drawArrowHeadLines(const gosiv_arrow&);
  void drawArrowHeadCurvy(const gosiv_arrow&);
  void drawLegendSolid(const gosiv_legend&);

  void transformArrowPoint(const gosiv_arrow&,gosiv_point&);
  void transformUser2Graph(double&,double&);
  void transformGraph2User(double&,double&);

  double text_width(const string&, int font);
  double text_height(const string&, int font);

  void drawImage(double,double,double,double,const gosiv_image&);
  void drawPath(const gosiv_color&,double,const gosiv_color&,gosiv_path&);
  void drawText(double,double,int,const gosiv_color&,double,int,int,const string&);

  void closePS();
  void drawImagePS(double,double,double,double,const gosiv_image&);
  void drawPathPS(const gosiv_color&,double,const gosiv_color&,gosiv_path&);
  void drawTextPS(double,double,int,const gosiv_color&,double,int,int,const string&);

  void closeSVG();
  void drawImageSVG(double,double,double,double,const gosiv_image&);
  void drawPathSVG(const gosiv_color&,double,const gosiv_color&,gosiv_path&);
  void drawTextSVG(double,double,int,const gosiv_color&,double,int,int,const string&);

};

#endif
