/* File: alg7.c 
 
 * Copyright (c) 2002 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
*/ 
 
#include "osiv.h" 
 
void OsivAlg7_CheckOpts (osiv_input* input) 
{ 
  if (input->PreProc == OSIV_IOINT_BLANK) OsivError("preprocessing method undefined"); 
  if (input->Interp == OSIV_IOINT_BLANK) OsivError("interpolation method undefined"); 
  if (input->XSize == OSIV_IOINT_BLANK) OsivError("image width undefined"); 
  if (input->YSize == OSIV_IOINT_BLANK) OsivError("image height undefined"); 
  if (input->XStart == OSIV_IOINT_BLANK) OsivError("x start position undefined"); 
  if (input->YStart == OSIV_IOINT_BLANK) OsivError("y start position undefined"); 
  if (input->XOffset == OSIV_IOINT_BLANK) OsivError("x offset undefined"); 
  if (input->YOffset == OSIV_IOINT_BLANK) OsivError("y offset undefined"); 
  if (input->XVectSpace == OSIV_IOINT_BLANK) OsivError("x spacing undefined"); 
  if (input->YVectSpace == OSIV_IOINT_BLANK) OsivError("y spacing undefined"); 
  if (input->XNumVects == OSIV_IOINT_BLANK) OsivError("number of x vectors undefined"); 
  if (input->YNumVects == OSIV_IOINT_BLANK) OsivError("number of y vectors undefined"); 
  if (input->XWinSize == OSIV_IOINT_BLANK) OsivError("x window size undefined"); 
  if (input->YWinSize == OSIV_IOINT_BLANK) OsivError("y window size undefined"); 
  if (input->XMaxDisp == OSIV_IOINT_BLANK) OsivError("maximum x displacement undefined"); 
  if (input->YMaxDisp == OSIV_IOINT_BLANK) OsivError("maximum y displacement undefined"); 
  if (input->XWinLgSize == OSIV_IOINT_BLANK) OsivError("super window x size undefined");  
  if (input->YWinLgSize == OSIV_IOINT_BLANK) OsivError("super window y size undefined");  
  if (input->FrmStart == OSIV_IOINT_BLANK) OsivError("start frame undefined"); 
  if (input->FrmFinish == OSIV_IOINT_BLANK) OsivError("stop frame undefined"); 
  if (input->FrmSkip == OSIV_IOINT_BLANK) OsivError("frame jump undefined"); 
  if (input->FrmSet == OSIV_IOINT_BLANK) OsivError("frame spacing undefined"); 
  if (!strlen(input->MovieFN)) OsivError("movie file undefined"); 
  if (!strlen(input->VectorsFN)) OsivError("vectors file undefined"); 
 
  if (input->PreProc < 0 || input->PreProc > 5) OsivError("unrecognized preprocessing method"); 
  if (input->Interp < 0 || input->Interp > 4) OsivError("unrecognized interpolation method"); 
  if (input->XSize < 1) OsivError("image width must be non-zero, non-negative"); 
  if (input->YSize < 1) OsivError("image height must be non-zero, non-negative"); 
  if (input->XStart < 1) OsivError("x start position must be non-negative"); 
  if (input->YStart < 1) OsivError("y start position must be non-negative"); 
  if (input->XVectSpace < 1) OsivError("x spacing must be non-zero, non-negative"); 
  if (input->YVectSpace < 1) OsivError("y spacing must be non-zero, non-negative"); 
  if (input->XNumVects < 1) OsivError("number of x vectors must be non-zero, non-negative"); 
  if (input->YNumVects < 1) OsivError("number of y vectors must be non-zero, non-negative"); 
  if (input->XWinSize < 1) OsivError("window x size must be non-zero, non-negative"); 
  if (input->YWinSize < 1) OsivError("window y size must be non-zero, non-negative"); 
  if (input->XMaxDisp < 1) OsivError("maximum x displacement must be non-zero, non-negative"); 
  if (input->YMaxDisp < 1) OsivError("maximum y displacement must be non-zero, non-negative"); 
  if (input->XWinLgSize < 1) OsivError("super window x size must be non-zero, non-negative");  
  if (input->YWinLgSize < 1) OsivError("super window y size must be non-zero, non-negative");  
  if (input->FrmStart < 0) OsivError("start frame must be non-negative"); 
  if (input->FrmFinish < 0) OsivError("stop frame must be non-negative"); 
  if (input->FrmSkip < 1) OsivError("frame jump must be non-zero, non-negative"); 
  if (input->FrmSet < 1) OsivError("frame spacing must be non-zero, non-negative"); 
 
  if ((log(2*input->XMaxDisp+1)+log(2*input->YMaxDisp+1))/log(2) >= OSIV_DISP_BITS) 
    OsivError("max displacement exceeds that specified at compile time"); 
  if ((log(2*(input->XWinSize/2+1))+log(input->YWinSize))/log(2) >= OSIV_WINDOW_BITS) 
    OsivError("window size exceeds that specified at compile time"); 
  if ((log(input->XSize)+log(input->YSize))/log(2) >= OSIV_FRAME_BITS) 
    OsivError("image size exceeds that specified at compile time"); 
  if (log(input->FrmFinish)/log(2) >= OSIV_MOVIE_BITS) 
    OsivError("movie length exceeds that specified at compile time"); 
 
  if (input->XStart + input->XOffset + 
      (input->XWinLgSize-input->XWinSize)/2 < 1) 
    OsivError("x start position is is too small"); 
  if (input->YStart + input->YOffset +  
      (input->YWinLgSize-input->YWinSize)/2 < 1) 
    OsivError("y start position is is too small"); 
  if (input->XSize < input->XStart +  
      (input->XNumVects-1)*input->XVectSpace + input->XWinLgSize) 
    OsivError("x interrogation domain exceeds limits of image"); 
  if (input->YSize < input->YStart +  
      (input->YNumVects-1)*input->YVectSpace + input->YWinLgSize) 
    OsivError("y interrogation domain exceeds limits of image"); 
  if (input->XSize < input->XStart + input->XOffset + 
      (input->XNumVects-1)*input->XVectSpace + 
      input->XWinSize+(input->XWinLgSize-input->XWinSize)/2) 
    OsivError("x interrogation domain exceeds limits of image"); 
  if (input->YSize < input->YStart + input->YOffset + 
      (input->YNumVects-1)*input->YVectSpace +  
      input->YWinSize+(input->YWinLgSize-input->YWinSize)/2) 
    OsivError("y interrogation domain exceeds limits of image"); 
  if (input->FrmFinish < input->FrmStart) 
    OsivError("last frame must follow first");   
  if (input->XWinLgSize < input->XWinSize+2*input->XMaxDisp) 
    OsivError("super window is too small"); 
  if (input->YWinLgSize < input->YWinSize+2*input->YMaxDisp) 
    OsivError("super window is too small"); 
} 
 
static void OsivAlg7_GetSquare (osiv_param* p, 
				osiv_imcp* im, 
				osiv_immf* out) 
{ 
  osiv_intf i,j,k; 
  osiv_intf xlen,ylen; 
  osiv_intf xmargin,ymargin; 
 
  xmargin = (p->XWinLgSize-p->XWinSize)/2; 
  ymargin = (p->YWinLgSize-p->YWinSize)/2; 
 
  xlen = (p->XNumVects-1)*p->XVectSpace+p->XWinSize+2*p->XMaxDisp+1; 
  ylen = (p->YNumVects-1)*p->YVectSpace+p->YWinSize+2*p->YMaxDisp+1; 
 
  /* Position image */ 
  im += xmargin + ymargin*p->XSize; 
  im -= (p->XMaxDisp+1)+(p->YMaxDisp+1)*p->XSize; 
 
  /* Set first pixel */ 
  *out = ((osiv_immp)*im)**im; 
 
  /* Integrate first row */ 
  for (i = 1; i < xlen; i++) 
    { 
      *(out+i) = *(out+i-1)+((osiv_immp)*(im+i))**(im+i); 
    } 
 
  /* Integrate rest of map */ 
  for (i = 1; i < ylen; i++) 
    { 
      *(out+i*xlen) = *(im+i*p->XSize)**(im+i*p->XSize); 
      for (j = 1+i*xlen, k = 1+i*p->XSize; j < xlen+i*xlen; j++,k++) 
	{ 
	  *(out+j) = *(out+j-1)+((osiv_immp)*(im+k))**(im+k); 
	  *(out+j-1) += *(out+j-1-xlen); 
	} 
      *(out+j-1) += *(out+j-1-xlen); 
    } 
 
  /* Evaluate integral for square */ 
  for (i = 0; i < ylen-p->YWinSize; i++) 
    { 
      for (j = i*xlen; j < xlen-p->XWinSize+i*xlen; j++) 
	{ 
	  *(out+j) += (osiv_immw)*(out+j+p->XWinSize+p->YWinSize*xlen) 
	    - *(out+j+p->XWinSize) - *(out+j+p->YWinSize*xlen); 
	} 
    } 
} 
 
static void OsivAlg7_Copyt0 (osiv_param* p, 
			     osiv_imcp* im, 
			     fftw_real* ft) 
{ 
  osiv_intf i,j; 
  osiv_intf xmargin,ymargin; 
 
  xmargin = (p->XWinLgSize-p->XWinSize)/2; 
  ymargin = (p->YWinLgSize-p->YWinSize)/2; 
 
  im += xmargin + ymargin*p->XSize; 
 
  for (i = 0; i < ymargin*p->XFFTLgSize; i++) *ft++ = 0; 
  for (i = 0; i < p->YWinSize*p->XSize; i+=p->XSize) 
    { 
      for (j = 0; j < xmargin; j++) *ft++ = 0; 
      for (j = i; j < p->XWinSize+i; j++) *ft++ = *(im+j); 
      for (j = xmargin+p->XWinSize; j < p->XFFTLgSize; j++) *ft++ = 0; 
    } 
  for (i = (ymargin+p->YWinSize)*p->XFFTLgSize; i < p->FFTLgLength; i++) *ft++ = 0; 
} 
 
static void OsivAlg7_Copyt1 (osiv_param* p, 
			     osiv_imcp* im, 
			     fftw_real* ft) 
{ 
  osiv_intf i,j; 
 
  for (i = 0; i < p->YWinLgSize*p->XSize; i+=p->XSize) 
    { 
      for (j = i; j < p->XWinLgSize+i; j++) *ft++ = *(im+j); 
      for (j = p->XWinLgSize; j < p->XFFTLgSize; j++) *ft++ = 0; 
    } 
} 
 
static void OsivAlg7_XCorr (fftw_real* a, 
			    fftw_real* b, 
			    osiv_intw len) 
{ 
  osiv_intw i; 
 
  for (i = 0; i < len; i += 2) 
    { 
      register fftw_real tmp; 
      tmp = 2*(*(a+i)**(b+i) + *(a+i+1)**(b+i+1))/len; 
      *(a+i+1) = 2*(*(a+i+1)**(b+i) - *(a+i)**(b+i+1))/len; 
      *(a+i) = tmp; 
    } 
} 
 
static void OsivAlg7_GetMQD (osiv_param* p, 
			     osiv_immf* imSq, 
			     fftw_real* out, 
			     osiv_intd* maxInd) 
{ 
  osiv_intf i,j,k; 
  osiv_intf sqdim; 
 
  sqdim= (p->XNumVects-1)*p->XVectSpace 
    +p->XWinSize+2*p->XMaxDisp+1; 
  imSq += p->XMaxDisp+p->YMaxDisp*sqdim; 
 
  *maxInd = k = 0; 
  for (i = 0; i <= p->YMaxDisp*sqdim; i+=sqdim) 
    { 
      for (j = i; j <= p->XMaxDisp+i; j++) 
	{ 
	  if (*(out+*maxInd) <= (*(out+k) = 2**(out+k)-*(imSq+j))) 
	    *maxInd = k; 
	  k++; 
	} 
      k += p->XWinLgSize-2*p->XMaxDisp-1; 
      for (j = -p->XMaxDisp+i; j < i; j++) 
	{ 
	  if (*(out+*maxInd) <= (*(out+k) = 2**(out+k)-*(imSq+j))) 
	    *maxInd = k; 
	  k++; 
	} 
      k += p->XFFTLgSize-p->XWinLgSize; 
    } 
 
  k += (p->YWinLgSize-2*p->YMaxDisp-1)*p->XFFTLgSize; 
  for (i = -p->YMaxDisp*sqdim; i < 0; i+=sqdim) 
    { 
      for (j = i; j <= p->XMaxDisp+i; j++) 
	{ 
	  if (*(out+*maxInd) <= (*(out+k) = 2**(out+k)-*(imSq+j))) 
	    *maxInd = k; 
	  k++; 
	} 
      k += p->XWinLgSize-2*p->XMaxDisp-1; 
      for (j = -p->XMaxDisp+i; j < i; j++) 
	{ 
	  if (*(out+*maxInd) <= (*(out+k) = 2**(out+k)-*(imSq+j))) 
	    *maxInd = k; 
	  k++; 
	} 
      k += p->XFFTLgSize-p->XWinLgSize; 
    } 
} 
 
static void OsivAlg7_ConvertMap (osiv_param* p, 
				 osiv_ioreal* dest, 
				 fftw_real* src) 
{ 
  osiv_intw i,j; 
 
  for (i = p->YWinLgSize-p->YMaxDisp; i < p->YWinLgSize; i++) 
    { 
      j = p->XWinLgSize-p->XMaxDisp; 
      for (j += i*p->XFFTLgSize; j < p->XWinLgSize+i*p->XFFTLgSize; j++) 
	*dest++ = *(src+j); 
 
      for (j = i*p->XFFTLgSize; j <= p->XMaxDisp+i*p->XFFTLgSize; j++) 
	*dest++ = *(src+j); 
    } 
 
  for (i = 0; i <= p->YMaxDisp; i++) 
    { 
      j = p->XWinLgSize-p->XMaxDisp; 
      for (j += i*p->XFFTLgSize; j < p->XWinLgSize+i*p->XFFTLgSize; j++) 
	*dest++ = *(src+j); 
 
      for (j = i*p->XFFTLgSize; j <= p->XMaxDisp+i*p->XFFTLgSize; j++) 
	*dest++ = *(src+j); 
    } 
} 
 
static void OsivAlg7_ConvertVects (osiv_param* p, 
				   osiv_ioreal* dest, 
				   osiv_real* src) 
{ 
  if (p->SwapReal) 
    { 
      osiv_intf i; 
      for (i = 0; i < 2*p->NumVects; i++) 
	*(dest+i) = (osiv_ioreal)*(src+i); 
    } 
} 
 
void OsivAlg7 (osiv_param *p) 
{ 
  osiv_intm i; 
  osiv_imcp* im0,*im1; 
  osiv_immf* imSq; 
  osiv_real* vects; 
  fftw_real* ft0,*ft1; 
  fftw_plan forw0,forw1; 
  fftw_plan back0; 
  osiv_ioreal* mapStd; 
  osiv_ioreal* vectsStd; 
 
  /* Allocate memory */ 
  im0 = (osiv_imcp*)OsivMalloc(p->ImageLength*sizeof(osiv_imcp)); 
  im1 = (osiv_imcp*)OsivMalloc(p->ImageLength*sizeof(osiv_imcp)); 
  imSq = (osiv_immf*)OsivMalloc(((p->XNumVects-1)*p->XVectSpace+p->XWinSize+2*p->XMaxDisp+1)* 
				((p->YNumVects-1)*p->YVectSpace+p->YWinSize+2*p->YMaxDisp+1)* 
				sizeof(osiv_immf)); 
  vects = (osiv_real*)OsivMalloc(2*p->NumVects*sizeof(osiv_real)); 
  ft0 = (fftw_real*)fftw_malloc(p->FFTLgLength*sizeof(fftw_real)); 
  ft1 = (fftw_real*)fftw_malloc(p->FFTLgLength*sizeof(fftw_real)); 
  if (!ft0 || !ft1) OsivError("unable to allocate memory"); 
 
  if (!p->WriteMap) 
      mapStd = NULL; 
  else 
      mapStd = (osiv_ioreal*)OsivMalloc(p->MapLength*sizeof(osiv_ioreal)); 
 
  if (!p->SwapReal) 
    vectsStd = (osiv_ioreal*)vects; 
  else 
    vectsStd = (osiv_ioreal*)OsivMalloc(2*p->NumVects*sizeof(osiv_ioreal)); 
 
  /* Generate fftw plans */ 
  forw0 = fftw_plan_dft_r2c_2d(p->YWinLgSize,p->XWinLgSize,ft0,(fftw_complex*)ft0,FFTW_ESTIMATE); 
  forw1 = fftw_plan_dft_r2c_2d(p->YWinLgSize,p->XWinLgSize,ft1,(fftw_complex*)ft1,FFTW_ESTIMATE); 
  back0 = fftw_plan_dft_c2r_2d(p->YWinLgSize,p->XWinLgSize,(fftw_complex*)ft0,ft0,FFTW_ESTIMATE); 
  if (!forw0 || !forw1 || !back0) OsivError("unable to complete fftw plan"); 
 
  /* Open body tag */ 
  OsivWriteTag(p->VectorsFile,OSIV_TAG_BODY); 
   
  /* Run through remainder of images */ 
  for (i = p->FrmStart; i <= p->FrmFinish; i += p->FrmSkip) 
    { 
      osiv_intf j,k; 
      osiv_imcp* im0Start,*im1Start; 
      osiv_real* vects_ptr = vects; 
      osiv_intf sqdim = (p->XNumVects-1)*p->XVectSpace 
	+p->XWinSize+2*p->XMaxDisp+1; 
 
      /* Read the ith image */ 
      if (p->FrmSkip != p->FrmSet || 
	  i == p->FrmStart) 
	{ 
	  OsivGetImage(p,i,im0); 
	  OsivPreProc(p,i,im0); 
	} 
      else 
	{ 
	  osiv_imcp* tmp = im1; 
	  im1 = im0; im0 = tmp; 
	} 
 
      /* Read the ith image pair */ 
      OsivGetImage(p,i+p->FrmSet,im1); 
      OsivPreProc(p,i+p->FrmSet,im1); 
 
      /* Open frame tag */ 
      OsivWriteTag(p->VectorsFile,OSIV_TAG_FRAME); 
 
      /* Write frame number */ 
      OsivWriteFrameNum(p->VectorsFile,i); 
 
      /* Set the start locations */ 
      im0Start = im0 + p->XStart + p->XOffset+ 
	(p->YStart+p->YOffset)*p->XSize; 
      im1Start = im1 + p->XStart + p->YStart*p->XSize; 
 
      /* Get square of im1Start */ 
      OsivAlg7_GetSquare(p,im0Start,imSq); 
 
      /* Run through each of the windows */ 
      for (j = 0; j < p->YNumVects; j++) 
	{ 
	  for (k = 0; k < p->XNumVects; k++) 
	    { 
	      osiv_intf impos,sqpos; 
	      osiv_intd ind,ix,iy; 
 
	      /* Set window position */ 
	      impos = k*p->XVectSpace  
		+j*p->YVectSpace*p->XSize; 
	      sqpos = k*p->XVectSpace  
		+j*p->YVectSpace*sqdim; 
 
	      /* Copy windows */ 
	      OsivAlg7_Copyt0(p,im0Start+impos,ft0); 
	      OsivAlg7_Copyt1(p,im1Start+impos,ft1); 
 
	      /* Calculate ffts */ 
	      fftw_execute(forw0); 
	      fftw_execute(forw1); 
 
	      /* Perform an in-place correlation */ 
	      OsivAlg7_XCorr(ft0,ft1,p->FFTLgLength); 
			 
	      /* Calculate the inverse fft */ 
	      fftw_execute(back0); 
	       
	      /* Add in square */ 
	      OsivAlg7_GetMQD(p,imSq+sqpos,ft0,&ind); 
	      ix = ind%p->XFFTLgSize; iy = ind/p->XFFTLgSize; 
			 
	      *vects_ptr = ix + p->XOffset 
		- p->XWinLgSize*(ix/(p->XWinLgSize/2+1)); 
	      *(vects_ptr+1) = iy + p->YOffset  
		- p->YWinLgSize*(iy/(p->YWinLgSize/2+1));       
 
	      /* Perform interpolation */ 
	      if (ix != p->XWinLgSize/2 && iy != p->YWinLgSize/2) 
		{ 
		  osiv_intd lft,rht,up,dwn; 
 
		  if (ix != 0) lft = -1; 
		  else lft = p->XWinLgSize-1; 
 
		  if (ix != p->XWinLgSize-1) rht = 1; 
		  else rht = -(p->XWinLgSize-1); 
 
		  if (iy != 0) up = -p->XFFTLgSize; 
		  else up = (p->YWinLgSize-1)*p->XFFTLgSize; 
 
		  if (iy != p->YWinLgSize-1) dwn = p->XFFTLgSize; 
		  else dwn = -(p->YWinLgSize-1)*p->XFFTLgSize; 
 
		  OsivInterpMax_FFTW(p->Interp, 
				    ft0+ind,  
				    lft,rht, 
				    up,dwn, 
				    vects_ptr,vects_ptr+1); 
		} 
 
	      /* Write error map to file */ 
	      if (p->WriteMap) 
		{ 
		  OsivAlg7_ConvertMap(p,mapStd,ft0); 
		  OsivWriteTag(p->VectorsFile,OSIV_TAG_BOX); 
		  OsivWriteBoxNum(p->VectorsFile,k+j*p->XNumVects); 
		  OsivWriteMap(p->VectorsFile,p->MapLength,mapStd); 
		  OsivWriteTag(p->VectorsFile,OSIV_GAT_BOX); 
		} 
 
	      vects_ptr += 2; 
	    } 
	} 
 
      /* Write vectors to file */ 
      OsivAlg7_ConvertVects(p,vectsStd,vects); 
      OsivWriteVectors(p->VectorsFile,2*p->NumVects,vectsStd); 
 
      /* Close frame tag */ 
      OsivWriteTag(p->VectorsFile,OSIV_GAT_FRAME); 
 
    } 
 
 
  /* Close body tag */ 
  OsivWriteTag(p->VectorsFile,OSIV_GAT_BODY); 
 
  /* Kill plan */ 
  fftw_destroy_plan(forw0); 
  fftw_destroy_plan(forw1); 
  fftw_destroy_plan(back0); 
 
  /* Free memory */ 
  free(im0); 
  free(im1); 
  free(vects); 
  fftw_free(ft0); 
  fftw_free(ft1); 
 
  if (p->WriteMap) 
    free(mapStd); 
  if (p->SwapReal) 
    free(vectsStd); 
} 
