/* File: save.c

 * Copyright (c) 2002 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include "osiv.h"


osiv_bool OsivWriteTag (FILE* fp,
			osiv_iotag tag)
{
  return (fwrite(&tag,sizeof(osiv_iotag),1,fp) != 1);
}


osiv_bool OsivWriteArgInt (FILE* fp,
			   osiv_ioint key,
			   osiv_ioint val)
{
  osiv_ioint blen = 2*sizeof(osiv_ioint);
  return (OsivWriteTag(fp,OSIV_TAG_ARG_INT) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&key,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&val,sizeof(osiv_ioint),1,fp) != 1 ||
	  OsivWriteTag(fp,OSIV_GAT_ARG_INT));
}

osiv_bool OsivWriteArgStr (FILE* fp,
			   osiv_ioint key,
			   char* str)
{
  osiv_ioint len = strlen(str);
  osiv_ioint blen = sizeof(osiv_ioint)+strlen(str)+1; 
  return (OsivWriteTag(fp,OSIV_TAG_ARG_STR) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&key,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(str,sizeof(char),len+1,fp) != len+1 ||
	  OsivWriteTag(fp,OSIV_GAT_ARG_STR));
}

osiv_bool OsivWriteFrameNum (FILE* fp,
			     osiv_ioint val)
{
  osiv_ioint blen = sizeof(osiv_ioint);
  return (OsivWriteTag(fp,OSIV_TAG_FRAME_NUM) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&val,sizeof(osiv_ioint),1,fp) != 1 ||
	  OsivWriteTag(fp,OSIV_GAT_FRAME_NUM));
}

osiv_bool OsivWriteVectors (FILE* fp,
			    osiv_ioint len,
			    osiv_ioreal* array)
{
  osiv_ioint blen = sizeof(osiv_ioint)+len*sizeof(osiv_ioreal);
  return (OsivWriteTag(fp,OSIV_TAG_VECTORS) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&len,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(array,sizeof(osiv_ioreal),len,fp) != len ||
	  OsivWriteTag(fp,OSIV_GAT_VECTORS));
}

osiv_bool OsivWriteBoxNum (FILE* fp,
			   osiv_ioint val)
{
  osiv_ioint blen = sizeof(osiv_ioint);
  return (OsivWriteTag(fp,OSIV_TAG_BOX_NUM) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&val,sizeof(osiv_ioint),1,fp) != 1 ||
	  OsivWriteTag(fp,OSIV_GAT_BOX_NUM));
}

osiv_bool OsivWriteMap (FILE* fp,
			osiv_ioint len,
			osiv_ioreal* map)
{
  osiv_ioint blen = sizeof(osiv_ioint)+len*sizeof(osiv_ioreal);
  return (OsivWriteTag(fp,OSIV_TAG_MAP) ||
	  fwrite(&blen,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(&len,sizeof(osiv_ioint),1,fp) != 1 ||
	  fwrite(map,sizeof(osiv_ioreal),len,fp) != len ||
	  OsivWriteTag(fp,OSIV_GAT_MAP));
}

osiv_bool OsivWriteHeader (FILE* fp,
			   osiv_input *input)
{
  osiv_uint16 unity = 1;

  /* Write pre-header */
  if (fputc((char)0xff,fp) == EOF ||
      fputc((char)0x00,fp) == EOF ||
      fwrite(&unity,sizeof(osiv_uint16),1,fp) != 1 ||
      fputc((char)sizeof(osiv_iotag),fp) == EOF ||
      fputc((char)sizeof(osiv_ioint),fp) == EOF ||
      fputc((char)sizeof(osiv_ioreal),fp) == EOF ||
      fputc((char)OSIV_VER_FIRST,fp) == EOF ||
      fputc((char)OSIV_VER_SECOND,fp) == EOF ||
      fputc((char)OSIV_VER_THIRD,fp) == EOF)
    return OSIV_TRUE;

  /* Open header tag */
  if (OsivWriteTag(fp,OSIV_TAG_HEADER))
    return OSIV_TRUE;

  /* Write arguments */
  if (OsivWriteArgInt(fp,OSIV_KEY_ALG,input->Alg) ||
      OsivWriteArgInt(fp,OSIV_KEY_PREPROC,input->PreProc) ||
      OsivWriteArgInt(fp,OSIV_KEY_INTERP,input->Interp) ||
      OsivWriteArgInt(fp,OSIV_KEY_XSIZE,input->XSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_YSIZE,input->YSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_XSTART,input->XStart) ||
      OsivWriteArgInt(fp,OSIV_KEY_YSTART,input->YStart) ||
      OsivWriteArgInt(fp,OSIV_KEY_XOFFSET,input->XOffset) ||
      OsivWriteArgInt(fp,OSIV_KEY_YOFFSET,input->YOffset) ||
      OsivWriteArgInt(fp,OSIV_KEY_XVECTSPACE,input->XVectSpace) ||
      OsivWriteArgInt(fp,OSIV_KEY_YVECTSPACE,input->YVectSpace) ||
      OsivWriteArgInt(fp,OSIV_KEY_XNUMVECTS,input->XNumVects) ||
      OsivWriteArgInt(fp,OSIV_KEY_YNUMVECTS,input->YNumVects) ||
      OsivWriteArgInt(fp,OSIV_KEY_XWINSIZE,input->XWinSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_YWINSIZE,input->YWinSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_XMAXDISP,input->XMaxDisp) ||
      OsivWriteArgInt(fp,OSIV_KEY_YMAXDISP,input->YMaxDisp) ||
      OsivWriteArgInt(fp,OSIV_KEY_XWINLGSIZE,input->XWinLgSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_YWINLGSIZE,input->YWinLgSize) ||
      OsivWriteArgInt(fp,OSIV_KEY_FRMSTART,input->FrmStart) ||
      OsivWriteArgInt(fp,OSIV_KEY_FRMFINISH,input->FrmFinish) ||
      OsivWriteArgInt(fp,OSIV_KEY_FRMSKIP,input->FrmSkip) ||
      OsivWriteArgInt(fp,OSIV_KEY_FRMSET,input->FrmSet) ||
      OsivWriteArgInt(fp,OSIV_KEY_NUMITER,input->NumIter) ||
      OsivWriteArgInt(fp,OSIV_KEY_WRITEMAP,input->WriteMap) ||
      OsivWriteArgStr(fp,OSIV_KEY_MOVIEFN,input->MovieFN) ||
      OsivWriteArgStr(fp,OSIV_KEY_VECTORSFN,input->VectorsFN))
    return OSIV_TRUE;

  /* Close header tag */
  if (OsivWriteTag(fp,OSIV_GAT_HEADER))
    return OSIV_TRUE;

  return OSIV_FALSE;
}
